from dataclasses import dataclass

from PIL import Image

from parse_qwantz.colors import Color, WHITE
from parse_qwantz.pixels import Pixel, get_pixels


@dataclass
class SimpleImage:
    width: int
    height: int
    pixels: dict[Pixel, Color]

    @classmethod
    def from_image(cls, image: Image, trim_top: bool = False):
        if trim_top:
            pixels = {pixel: color for pixel, color in get_pixels(image) if pixel.x > 240 or pixel.y > 46}
        else:
            pixels = dict(get_pixels(image))
        return cls(image.width, image.height, pixels)

    def get_pixel(self, pixel: Pixel) -> Color:
        # caution: no bounds checking!
        return self.pixels.get(pixel, WHITE)

    def is_on_edge(self, pixel: Pixel) -> bool:
        x, y = pixel
        return x in (0, self.width - 1) or y in (0, self.height - 1)

    def distance_to_edge(self, pixel: Pixel) -> int:
        return min(pixel.x, pixel.y, self.width - pixel.x - 1, self.height - pixel.y - 1)



import subprocess

from PIL.ImageShow import XDGViewer


class SilentViewer(XDGViewer):
    def show_file(self, path, **options):
        """
        Display given file.
        """
        subprocess.call(["xdg-open", path], stderr=subprocess.DEVNULL)
        return 1



from typing import Optional, NamedTuple

COLOR_THRESHOLD = 300


class Color(NamedTuple):
    red: int
    green: int
    blue: int


WHITE = Color(255, 255, 255)
BLACK = Color(0, 0, 0)
RED = Color(128, 0, 0)
GREY = Color(128, 128, 128)
GREY2 = Color(117, 117, 117)
GREY3 = Color(100, 100, 100)


def get_color_with_threshold(rgb: tuple[int, int, int]) -> Optional["Color"]:
    for color in [WHITE, BLACK, RED, GREY, GREY2, GREY3]:
        if square_distance(color, rgb) < COLOR_THRESHOLD:
            return color


def square_distance(rgb1: tuple[int, int, int], rgb2: tuple[int, int, int]):
    r1, g1, b1 = rgb1
    r2, g2, b2 = rgb2
    return (r1 - r2)**2 + (g1 - g2)**2 + (b1 - b2)**2



from typing import Collection

from parse_qwantz.box import Box
from parse_qwantz.colors import Color
from parse_qwantz.pixels import Pixel
from parse_qwantz.simple_image import SimpleImage


def get_box(shape: Collection[Pixel], padding=0) -> Box:
    x_min = min(x for x, y in shape)
    x_max = max(x for x, y in shape)
    y_min = min(y for x, y in shape)
    y_max = max(y for x, y in shape)
    return Box(Pixel(x_min - padding, y_min - padding), Pixel(x_max + 1 + padding, y_max + 1 + padding))


def get_shape(pixel: Pixel, image: SimpleImage) -> dict[Pixel, Color]:
    pixels_left = [pixel]
    pixels = {}
    while pixels_left:
        pixel = pixels_left.pop()
        pixels[pixel] = image.get_pixel(pixel)
        for other_pixel in get_neighbor_pixels(pixel):
            if other_pixel not in pixels and other_pixel in image.pixels:
                pixels_left.append(other_pixel)
    return pixels


def get_adjacent_pixels(pixel: Pixel) -> list[Pixel]:
    x, y = pixel
    return [
        Pixel(x - 1, y),
        Pixel(x, y - 1),
        Pixel(x, y + 1),
        Pixel(x + 1, y),
    ]


def get_neighbor_pixels(pixel: Pixel) -> list[Pixel]:
    x, y = pixel
    return [
        Pixel(x - 1, y - 1),
        Pixel(x - 1, y),
        Pixel(x - 1, y + 1),
        Pixel(x, y - 1),
        Pixel(x, y + 1),
        Pixel(x + 1, y - 1),
        Pixel(x + 1, y),
        Pixel(x + 1, y + 1),
    ]



from collections import deque
from itertools import product

from parse_qwantz.box import Box
from parse_qwantz.pixels import Pixel
from parse_qwantz.simple_image import SimpleImage
from parse_qwantz.shape import get_shape, get_adjacent_pixels, get_box


def get_thought(pixel: Pixel, image: SimpleImage) -> tuple[Box, list[Pixel]] | None:
    orig_pixels = get_shape(pixel, image)
    box = get_box(orig_pixels)
    if any(x == 0 for x, y in orig_pixels) and any(x == image.width - 1 for x, y in orig_pixels):
        if len(set(y for x, y in orig_pixels)) > 2:
            box = Box(Pixel(0, 0), box.bottom_right)
            return box, sorted(orig_pixels)
        else:
            return None
    tripled_pixels: set[Pixel] = set(orig_pixels)
    for x, y in orig_pixels:
        tripled_pixels.update({Pixel(x + 1, y), Pixel(x + 2, y)})
    if is_thought(tripled_pixels, box, image) or is_thought(set(orig_pixels), box, image):
        return box, sorted(orig_pixels)


def is_thought(pixels: set[Pixel], box: Box, image: SimpleImage) -> bool:
    x_range = range(box.left, box.right)
    y_range = range(box.top, box.bottom)
    outside = set()
    outside.update(Pixel(box.left, y) for y in y_range)
    outside.update(Pixel(box.right - 1, y) for y in y_range)
    outside.update((Pixel(x, box.top) for x in x_range))
    outside.update((Pixel(x, box.bottom - 1) for x in x_range))
    for x, y in product(x_range, y_range):
        if Pixel(x, y) in pixels:
            continue
        visited = set()
        to_visit = deque([Pixel(x, y)])
        while to_visit:
            current = to_visit.popleft()
            on_image_border = image.is_on_edge(current)
            if current in visited or current in pixels or not box.includes(current) or on_image_border:
                continue
            if current in outside:
                outside.update(visited)
                break
            visited.add(current)
            to_visit.extend(get_adjacent_pixels(current))
        else:
            return True
    return False
